"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from ssl import SSLContext
from time import sleep as block_sleep
from typing import Iterator
from typing import Optional
from typing import Protocol
from typing import overload
from unittest.mock import MagicMock
from unittest.mock import Mock

from encommon.types import DictStrAny

from httpx import Response

from pytest import fixture

from pytest_mock import MockerFixture

from respx import MockRouter



_EVENTS = Optional[list[DictStrAny]]

_SOCKET = tuple[
    SSLContext,
    MagicMock]



EVENTS: list[DictStrAny] = [

    {'event': 'posted',
     'seq': 4,
     'broadcast': {
         'channel_id': 'chanid'},
     'data': {
         'channel_type': 'D',
         'post': (
             '{"user_id":"userid",'
             '"channel_id":"privid",'
             '"message":"Hello mtmbot"}'),
         'sender_name': '@user'}},

    {'event': 'posted',
     'seq': 5,
     'broadcast': {
         'channel_id': 'chanid'},
     'data': {
         'channel_type': 'P',
         'post': (
             '{"user_id":"userid",'
             '"channel_id":"chanid",'
             '"message":"Hello world"}'),
         'sender_name': '@user'}},

    {'event': 'posted',
     'seq': 6,
     'broadcast': {
         'channel_id': 'chanid'},
     'data': {
         'channel_type': 'P',
         'post': (
             '{"user_id":"mtmunq",'
             '"channel_id":"chanid",'
             '"message":"Hello user"}'),
         'sender_name': '@mtmbot'}}]



RVENTS: list[DictStrAny] = [

    {'status': 'OK',
     'seq_reply': 1},

    {'event': 'hello',
     'seq': 0},

    {'event': 'status_change',
     'broadcast': {
         'user_id': 'mtmunq'},
     'data': {
         'status': 'online',
         'user_id': 'mtmunq'},
     'seq': 1},

    {'status': 'OK',
     'seq_reply': 2},

    {'status': 'OK',
     'seq_reply': 3}]



WHOME: DictStrAny = {
    'id': 'mtmunq',
    'username': 'mtmbot'}



class MTMClientSocket(Protocol):
    """
    Typing protocol which the developer does not understand.
    """

    @overload
    def __call__(
        self,
        rvents: _EVENTS,
    ) -> _SOCKET:
        ...  # NOCVR

    @overload
    def __call__(
        self,
    ) -> _SOCKET:
        ...  # NOCVR

    def __call__(
        self,
        rvents: _EVENTS = None,
    ) -> _SOCKET:
        """
        Construct the instance for use in the downstream tests.

        :param rvents: Raw events for playback from the server.
        """
        ...  # NOCVR



@fixture
def client_mtmsock(  # noqa: CFQ004
    mocker: MockerFixture,
    respx_mock: MockRouter,
) -> MTMClientSocket:
    """
    Construct the instance for use in the downstream tests.

    :param mocker: Object for mocking the Python routines.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    content = dumps(WHOME)

    (respx_mock
     .get(
         'https://mocked:443'
         '/api/v4/users/me')
     .mock(Response(
         status_code=200,
         content=content)))

    (respx_mock
     .post(
         'https://mocked:443'
         '/api/v4/posts')
     .mock(Response(200)))


    socmod = mocker.patch(
        ('enconnect.mattermost'
         '.client.connect'),
        autospec=True)


    def _encode(
        resps: list[DictStrAny],
    ) -> list[bytes]:

        return [
            dumps(x).encode('utf-8')
            for x in resps]


    def _delayed(
        events: list[bytes],
    ) -> Iterator[bytes]:

        while True:

            for event in events:

                block_sleep(0.1)

                yield event

            block_sleep(0.1)

            # This event is not using
            # one that actually exist
            tneve = {
                'event': 'discon',
                'seq': 69420,
                'error': {
                    'reason': 'EOF'}}

            yield (
                dumps(tneve)
                .encode('utf-8'))


    def _factory(
        rvents: list[DictStrAny],
    ) -> MagicMock:

        effect = _delayed(
            _encode(rvents))

        socket = MagicMock()

        socket.send = Mock()

        socket.recv = Mock(
            side_effect=effect)

        socket.close = Mock()

        return socket


    def _fixture(
        rvents: _EVENTS = None,
    ) -> _SOCKET:

        rvents = rvents or []

        socket = _factory(
            RVENTS + rvents)

        socmod.return_value = socket

        return (socmod, socket)


    return _fixture



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import raises

from .helpers import EVENTS
from ..client import Client
from ..models import ClientEvent
from ..params import ClientParams
from ...fixtures import MTMClientSocket



def test_ClientEvent() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    params = ClientParams(
        server='mocked',
        token='mocked',
        teamid='mocked')

    client = Client(params)


    _event = {
        'status': 'OK',
        'seq_reply': 1}

    event = ClientEvent(
        client, _event)


    attrs = lattrs(event)

    assert attrs == [
        'type',
        'data',
        'broadcast',
        'seqno',
        'status',
        'error',
        'seqre',
        'original',
        'kind',
        'isme',
        'hasme',
        'whome',
        'author',
        'recipient',
        'message']


    assert inrepr(
        'ClientEvent(type',
        event)

    with raises(TypeError):
        hash(event)

    assert instr(
        'type=None data=None',
        event)


    assert event.original

    assert event.kind == 'event'

    assert not event.isme

    assert not event.hasme

    assert not event.whome

    assert not event.author

    assert not event.recipient

    assert not event.message



def test_ClientEvent_cover(  # noqa: CFQ001
    client_mtmsock: MTMClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client_ircsock: Object to mock client connection.
    """

    params = ClientParams(
        server='mocked',
        token='mocked',
        teamid='mocked')

    client = Client(params)


    def _operate() -> None:

        client_mtmsock(EVENTS)

        _raises = ConnectionError

        with raises(_raises):
            client.operate()


    thread = Thread(
        target=_operate)

    thread.start()


    mqueue = client.mqueue


    item = mqueue.get()

    assert not item.type
    assert not item.data
    assert not item.broadcast
    assert not item.seqno
    assert item.status == 'OK'
    assert not item.error
    assert item.seqre == 1

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message

    assert not client.canceled
    assert client.connected
    assert client.nickname == (
        'mtmbot', 'mtmunq')


    item = mqueue.get()

    assert item.type == 'hello'
    assert not item.data
    assert not item.broadcast
    assert not item.seqno
    assert not item.status
    assert not item.error
    assert not item.seqre

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert item.type == (
        'status_change')
    assert item.data
    assert item.broadcast
    assert item.seqno == 1
    assert not item.status
    assert not item.error
    assert not item.seqre

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert not item.type
    assert not item.data
    assert not item.broadcast
    assert not item.seqno
    assert item.status == 'OK'
    assert not item.error
    assert item.seqre == 2

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert not item.type
    assert not item.data
    assert not item.broadcast
    assert not item.seqno
    assert item.status == 'OK'
    assert not item.error
    assert item.seqre == 3

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert item.type == 'posted'
    assert item.data
    assert len(item.data) == 3
    assert item.broadcast
    assert len(item.broadcast) == 1
    assert item.seqno == 4
    assert not item.status
    assert not item.error
    assert not item.seqre

    assert item.kind == 'privmsg'
    assert not item.isme
    assert item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert item.author == (
        'user', 'userid')
    assert item.recipient == (
        'privid')
    assert item.message == (
        'Hello mtmbot')


    item = mqueue.get()

    assert item.type == 'posted'
    assert item.data
    assert len(item.data) == 3
    assert item.broadcast
    assert len(item.broadcast) == 1
    assert item.seqno == 5
    assert not item.status
    assert not item.error
    assert not item.seqre

    assert item.kind == 'chanmsg'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert item.author == (
        'user', 'userid')
    assert item.recipient == (
        'chanid')
    assert item.message == (
        'Hello world')


    item = mqueue.get()

    assert item.type == 'posted'
    assert item.data
    assert len(item.data) == 3
    assert item.broadcast
    assert len(item.broadcast) == 1
    assert item.seqno == 6
    assert not item.status
    assert not item.error
    assert not item.seqre

    assert item.kind == 'chanmsg'
    assert item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert item.author == (
        'mtmbot', 'mtmunq')
    assert item.recipient == (
        'chanid')
    assert item.message == (
        'Hello user')


    item = mqueue.get()

    assert item.type == 'discon'
    assert not item.data
    assert not item.broadcast
    assert item.seqno == 69420
    assert not item.status
    assert item.error
    assert len(item.error) == 1
    assert not item.seqre

    assert item.kind == 'event'
    assert not item.isme
    assert not item.hasme
    assert item.whome
    assert item.whome[0] == 'mtmbot'
    assert not item.author
    assert not item.recipient
    assert not item.message

    assert not client.canceled
    assert not client.connected
    assert client.nickname == (
        'mtmbot', 'mtmunq')


    thread.join(10)


    assert mqueue.empty()



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .helpers import DSCClientSocket
from .helpers import EVENTS
from .helpers import RVENTS
from .helpers import client_dscsock



__all__ = [
    'DSCClientSocket',
    'client_dscsock',
    'EVENTS',
    'RVENTS']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .bridge import Bridge
from .params import BridgeParams



__all__ = [
    'Bridge',
    'BridgeParams']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .client import Client
from .models import ClientEvent
from .params import ClientParams



__all__ = [
    'Client',
    'ClientParams',
    'ClientEvent']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from json import loads

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture
from pytest import mark

from respx import MockRouter

from . import SAMPLES
from ..params import RedditParams
from ..reddit import Reddit



@fixture
def social() -> Reddit:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = RedditParams(
        username='mocked',
        password='mocked',
        client='mocked',
        secret='mocked',
        useragent='mocked')

    return Reddit(params)



def test_Reddit(
    social: Reddit,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    """


    attrs = lattrs(social)

    assert attrs == [
        '_Reddit__params',
        '_Reddit__client',
        '_Reddit__token']


    assert inrepr(
        'reddit.Reddit object',
        social)

    assert isinstance(
        hash(social), int)

    assert instr(
        'reddit.Reddit object',
        social)


    assert social.params

    assert social.client

    assert not social.token



def test_Reddit_latest_block(
    social: Reddit,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _token = read_text(
        SAMPLES / 'token.json')

    location = [
        'https://oauth.reddit.com',
        'https://www.reddit.com']


    (respx_mock
     .post(
         f'{location[1]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .post(
         f'{location[0]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .get(
         f'{location[0]}/r'
         '/mocked/new.json')
     .mock(side_effect=[
         Response(
             status_code=401),
         Response(
             status_code=200,
             content=_latest)]))


    latest = (
        social.latest_block('mocked'))


    sample_path = (
        SAMPLES / 'latest.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in latest],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in latest])

    assert expect == sample


    social.request_token_block()



@mark.asyncio
async def test_Reddit_latest_async(
    social: Reddit,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _token = read_text(
        SAMPLES / 'token.json')

    location = [
        'https://oauth.reddit.com',
        'https://www.reddit.com']


    (respx_mock
     .post(
         f'{location[1]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .post(
         f'{location[0]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .get(
         f'{location[0]}/r'
         '/mocked/new.json')
     .mock(side_effect=[
         Response(
             status_code=401),
         Response(
             status_code=200,
             content=_latest)]))


    latest = await (
        social.latest_async('mocked'))


    sample_path = (
        SAMPLES / 'latest.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in latest],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in latest])

    assert expect == sample


    await social.request_token_async()



def test_Reddit_listing_block(
    social: Reddit,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _token = read_text(
        SAMPLES / 'token.json')

    _listing = dumps([
        loads(_latest)])

    location = [
        'https://oauth.reddit.com',
        'https://www.reddit.com']


    (respx_mock
     .post(
         f'{location[1]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .post(
         f'{location[0]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .get(
         f'{location[0]}'
         '/comments/mocked.json')
     .mock(side_effect=[
         Response(
             status_code=401),
         Response(
             status_code=200,
             content=_listing)]))


    listing = (
        social.listing_block('mocked'))


    sample_path = (
        SAMPLES / 'listing.json')

    sample = load_sample(
        sample_path,
        listing.endumped,
        update=ENPYRWS)

    expect = prep_sample(
        listing.endumped)

    assert expect == sample



@mark.asyncio
async def test_Reddit_listing_async(
    social: Reddit,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _token = read_text(
        SAMPLES / 'token.json')

    _listing = dumps([
        loads(_latest)])

    location = [
        'https://oauth.reddit.com',
        'https://www.reddit.com']


    (respx_mock
     .post(
         f'{location[1]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .post(
         f'{location[0]}/api'
         '/v1/access_token')
     .mock(Response(
         status_code=200,
         content=_token)))

    (respx_mock
     .get(
         f'{location[0]}'
         '/comments/mocked.json')
     .mock(side_effect=[
         Response(
             status_code=401),
         Response(
             status_code=200,
             content=_listing)]))


    listing = await (
        social.listing_async('mocked'))


    sample_path = (
        SAMPLES / 'listing.json')

    sample = load_sample(
        sample_path,
        listing.endumped,
        update=ENPYRWS)

    expect = prep_sample(
        listing.endumped)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import read_text

from pytest import raises

from . import SAMPLES
from ..models import YouTubeResult
from ..models import YouTubeVideo



def test_YouTubeResult() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    content = read_text(
        f'{SAMPLES}/search'
        '/source.json')

    loaded = loads(content)

    assert isinstance(
        loaded, dict)

    source = loaded['items'][0]


    item = (
        YouTubeResult(**source))


    attrs = lattrs(item)

    assert attrs == [
        'kind',
        'channel',
        'playlist',
        'video',
        'title',
        'about',
        'channel_title',
        'thumbnail',
        'published']


    assert inrepr(
        'YouTubeResult(kind',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        "kind='video'",
        item)

    assert item.kind == 'video'

    assert item.channel == 'UCPNB7Hrq5oHl81EzG6x3BwA'

    assert item.playlist is None

    assert item.video == 'FIEkHYHDjzk'

    assert item.title == 'Validating encommon code'

    assert item.about
    assert item.about[:12] == 'Demonstrates'

    assert item.channel_title == 'Enasis Network'

    assert item.thumbnail[:5] == 'https'

    assert item.published[:4] == '2024'




def test_YouTubeVideo() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    content = read_text(
        f'{SAMPLES}/videos'
        '/source.json')

    loaded = loads(content)

    assert isinstance(
        loaded, dict)

    source = loaded['items'][0]


    item = (
        YouTubeVideo(**source))


    attrs = lattrs(item)

    assert attrs == [
        'kind',
        'channel',
        'video',
        'title',
        'about',
        'thumbnail',
        'published']


    assert inrepr(
        'YouTubeVideo(kind',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        "kind='video'",
        item)


    assert item.kind == 'video'

    assert item.channel == 'UCPNB7Hrq5oHl81EzG6x3BwA'

    assert item.video == '9h-vcAjRCz8'

    assert item.title == 'Validating encommon code'

    assert item.about
    assert item.about[:12] == 'Demonstrates'

    assert item.thumbnail[:5] == 'https'

    assert item.published[:4] == '2024'



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from copy import deepcopy
from ssl import SSLContext
from time import sleep
from typing import AsyncIterator
from typing import Iterator
from typing import Literal
from typing import Optional

from encommon.types import DictStrAny

from httpx import AsyncClient
from httpx import Client as BlockClient
from httpx import Response
from httpx._client import UseClientDefault



_METHODS = Literal[
    'delete',
    'get',
    'post',
    'patch',
    'put']

_HTTPAUTH = tuple[str, str]

_HEADERS = dict[str, str]

_PAYLOAD = DictStrAny

_VERIFY = SSLContext | str | bool



class HTTPClient:
    """
    Interact with the upstream server in blocking or async.

    :param timeout: Timeout waiting for the server response.
    :param headers: Optional headers to include in requests.
    :param verify: Require valid certificate from the server.
    :param capem: Optional path to the certificate authority.
    :param httpauth: Optional information for authentication.
    :param retry: How many attempts are made with the server.
    :param backoff: Backoff backoff if encountered retries.
    :param states: Which states will be retried with backoff.
    """

    __timeout: int
    __headers: Optional[_HEADERS]
    __verify: _VERIFY
    __capem: Optional[str]
    __httpauth: Optional[_HTTPAUTH]
    __retry: int
    __backoff: float
    __states: set[int]

    __client_block: BlockClient
    __client_async: AsyncClient


    def __init__(  # noqa: CFQ002
        self,
        timeout: int = 30,
        headers: Optional[_HEADERS] = None,
        verify: _VERIFY = True,
        capem: Optional[str] = None,
        httpauth: Optional[_HTTPAUTH] = None,
        retry: int = 3,
        backoff: float = 3.0,
        states: set[int] = {429},
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        timeout = int(timeout)
        headers = deepcopy(headers)
        httpauth = deepcopy(httpauth)
        retry = int(retry)
        backoff = float(backoff)
        states = set(states)

        client_block = BlockClient(
            timeout=timeout,
            headers=headers or None,
            auth=httpauth or None,
            verify=capem or verify,
            follow_redirects=True)

        client_async = AsyncClient(
            timeout=timeout,
            headers=headers or None,
            auth=httpauth or None,
            verify=capem or verify,
            follow_redirects=True)

        self.__timeout = timeout
        self.__headers = headers
        self.__verify = verify
        self.__capem = capem
        self.__httpauth = httpauth
        self.__retry = retry
        self.__backoff = backoff
        self.__states = states

        self.__client_block = client_block
        self.__client_async = client_async


    @property
    def timeout(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__timeout


    @property
    def httpauth(
        self,
    ) -> Optional[_HTTPAUTH]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return deepcopy(self.__httpauth)


    @property
    def headers(
        self,
    ) -> Optional[_HEADERS]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return deepcopy(self.__headers)


    @property
    def verify(
        self,
    ) -> _VERIFY:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__verify


    @property
    def capem(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__capem


    @property
    def retry(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__retry


    @property
    def backoff(
        self,
    ) -> float:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__backoff


    @property
    def states(
        self,
    ) -> set[int]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__states


    @property
    def client_block(
        self,
    ) -> BlockClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client_block


    @property
    def client_async(
        self,
    ) -> AsyncClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client_async


    def request_block(  # noqa: CFQ002
        self,
        method: _METHODS,
        location: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        data: Optional[_PAYLOAD] = None,
        timeout: Optional[int] = None,
        headers: Optional[_HEADERS] = None,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param location: Location with path for server request.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param data: Optional dict payload included in request.
        :param timeout: Timeout waiting for the server response.
        :param headers: Optional headers to include in requests.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        retry = self.__retry
        backoff = self.__backoff
        states = self.__states

        default = UseClientDefault()

        client = self.__client_block
        request = client.request


        for count in range(retry):

            response = request(
                method=method,
                url=location,
                headers=headers or None,
                auth=httpauth or default,
                timeout=timeout or default,
                params=params or None,
                data=data or None,
                json=json or None)

            status = response.status_code

            if status not in states:
                break

            sleep(backoff)


        return response


    async def request_async(  # noqa: CFQ002
        self,
        method: _METHODS,
        location: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        data: Optional[_PAYLOAD] = None,
        timeout: Optional[int] = None,
        headers: Optional[_HEADERS] = None,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param location: Location with path for server request.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param data: Optional dict payload included in request.
        :param timeout: Timeout waiting for the server response.
        :param headers: Optional headers to include in requests.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        retry = self.__retry
        backoff = self.__backoff
        states = self.__states

        default = UseClientDefault()

        client = self.__client_async
        request = client.request


        for count in range(retry):

            response = await request(
                method=method,
                url=location,
                headers=headers or None,
                auth=httpauth or default,
                timeout=timeout or default,
                params=params or None,
                data=data or None,
                json=json or None)

            status = response.status_code

            if status not in states:
                break

            await asyncio.sleep(backoff)


        await asyncio.sleep(0)

        return response


    def stream_block(  # noqa: CFQ002
        self,
        method: _METHODS,
        location: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        data: Optional[_PAYLOAD] = None,
        timeout: Optional[int] = None,
        headers: Optional[_HEADERS] = None,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> Iterator[str]:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param location: Location with path for server request.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param data: Optional dict payload included in request.
        :param timeout: Timeout waiting for the server response.
        :param headers: Optional headers to include in requests.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        default = UseClientDefault()

        client = self.__client_block
        request = client.stream


        stream = request(
            method=method,
            url=location,
            headers=headers or None,
            auth=httpauth or default,
            timeout=timeout or default,
            params=params or None,
            data=data or None,
            json=json or None)


        with stream as _stream:

            lines = (
                _stream.iter_lines())

            yield from lines


    async def stream_async(  # noqa: CFQ002
        self,
        method: _METHODS,
        location: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        data: Optional[_PAYLOAD] = None,
        timeout: Optional[int] = None,
        headers: Optional[_HEADERS] = None,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> AsyncIterator[str]:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param location: Location with path for server request.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param data: Optional dict payload included in request.
        :param timeout: Timeout waiting for the server response.
        :param headers: Optional headers to include in requests.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        default = UseClientDefault()

        client = self.__client_async
        request = client.stream


        stream = request(
            method=method,
            url=location,
            headers=headers or None,
            auth=httpauth or default,
            timeout=timeout or default,
            params=params or None,
            data=data or None,
            json=json or None)


        async with stream as _stream:

            lines = (
                _stream.aiter_lines())

            async for line in lines:
                yield line



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from json import dumps
from json import loads
from typing import AsyncIterator
from typing import Iterator

from encommon.types import DictStrAny

from httpx import AsyncByteStream
from httpx import SyncByteStream



class ByteStreamBlock(SyncByteStream):
    """
    Class used for gaining coverage in dependency projects.

    :param source: JSON encoded source string for iteration.
    """

    source: list[DictStrAny]

    def __init__(
        self,
        source: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.source = loads(source)

    def __iter__(
        self,
    ) -> Iterator[bytes]:
        """
        Return the iterator for enumerating the provided events.
        """

        source = self.source

        chunks = [
            (f'data: {dumps(x)}\n'
             .encode('utf-8'))
            for x in source]

        chunks.insert(0, b': hi\n')

        yield from chunks



class ByteStreamAsync(AsyncByteStream):
    """
    Class used for gaining coverage in dependency projects.

    :param source: JSON encoded source string for iteration.
    """

    source: list[DictStrAny]

    def __init__(
        self,
        source: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.source = loads(source)


    async def __aiter__(
        self,
    ) -> AsyncIterator[bytes]:
        """
        Return the iterator for enumerating the provided events.
        """

        source = self.source

        chunks = [
            (f'data: {dumps(x)}\n'
             .encode('utf-8'))
            for x in source]

        chunks.insert(0, b': hi\n')

        await asyncio.sleep(0)

        for chunk in chunks:

            yield chunk

            await asyncio.sleep(0)

        await asyncio.sleep(0)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Literal
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field



RESULT_KINDS = Literal[
    'channel',
    'playlist',
    'video']

RESULT_VALUE = {
    'channelId': 'channel',
    'channelTitle': 'channel_title',
    'description': 'about',
    'publishedAt': 'published',
    'thumbnails': 'thumbnail',
    'title': 'title'}

RESULT_BASIC = {
    'channelId': 'channel',
    'kind': 'kind',
    'playlistId': 'playlist',
    'videoId': 'video'}

VIDEO_VALUE = {
    'channelId': 'channel',
    'channelTitle': 'channel_title',
    'description': 'about',
    'publishedAt': 'published',
    'thumbnails': 'thumbnail',
    'title': 'title'}



class YouTubeResult(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream response.
    """

    kind: Annotated[
        RESULT_KINDS,
        Field(...,
              description='Value from the server response')]

    channel: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    playlist: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    video: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    title: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    about: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    channel_title: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    thumbnail: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    published: Annotated[
        str,
        Field(...,
              description='Value from the server response')]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        basic = data.get('id', {})

        items = RESULT_BASIC.items()

        for old, new in items:

            value = basic.get(old)

            if value is None:
                continue

            data[new] = value


        match = data.get('snippet', {})

        items = RESULT_VALUE.items()

        for old, new in items:

            value = match.get(old)

            if value is None:
                continue  # NOCVR

            data[new] = value


        if 'thumbnail' in data:
            data['thumbnail'] = (
                data['thumbnail']
                ['high']['url'])


        data['kind'] = data['kind'][8:]


        super().__init__(**data)



class YouTubeVideo(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream response.
    """

    kind: Annotated[
        RESULT_KINDS,
        Field(...,
              description='Value from the server response')]

    channel: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    video: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    title: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    about: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    thumbnail: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    published: Annotated[
        str,
        Field(...,
              description='Value from the server response')]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        data['video'] = data['id']


        video = data['snippet']

        items = VIDEO_VALUE.items()

        for old, new in items:

            value = video.get(old)

            if value is None:
                continue  # NOCVR

            data[new] = value


        if 'thumbnail' in data:
            data['thumbnail'] = (
                data['thumbnail']
                ['high']['url'])


        data['kind'] = data['kind'][8:]


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture
from pytest import mark

from respx import MockRouter

from . import SAMPLES
from ..params import YouTubeParams
from ..youtube import YouTube



@fixture
def social() -> YouTube:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = YouTubeParams(
        token='mocked')

    return YouTube(params)



def test_YouTube(
    social: YouTube,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    """


    attrs = lattrs(social)

    assert attrs == [
        '_YouTube__params',
        '_YouTube__client']


    assert inrepr(
        'youtube.YouTube object',
        social)

    assert isinstance(
        hash(social), int)

    assert instr(
        'youtube.YouTube object',
        social)


    assert social.params

    assert social.client



def test_YouTube_search_block(
    social: YouTube,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _search = read_text(
        f'{SAMPLES}/search'
        '/source.json')

    location = (
        'https://www.googleapis.com'
        '/youtube/v3')


    (respx_mock
     .get(f'{location}/search')
     .mock(Response(
         status_code=200,
         content=_search)))


    search = (
        social.search_block(
            {'channelId': 'mocked'}))


    sample_path = (
        f'{SAMPLES}/search'
        '/dumped.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in search],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in search])

    assert expect == sample



@mark.asyncio
async def test_YouTube_search_async(
    social: YouTube,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _search = read_text(
        f'{SAMPLES}/search'
        '/source.json')

    location = (
        'https://www.googleapis.com'
        '/youtube/v3')


    (respx_mock
     .get(f'{location}/search')
     .mock(Response(
         status_code=200,
         content=_search)))


    search = await (
        social.search_async(
            {'channelId': 'mocked'}))


    sample_path = (
        f'{SAMPLES}/search'
        '/dumped.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in search],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in search])

    assert expect == sample



def test_YouTube_videos_block(
    social: YouTube,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _videos = read_text(
        f'{SAMPLES}/videos'
        '/source.json')

    location = (
        'https://www.googleapis.com'
        '/youtube/v3')


    (respx_mock
     .get(f'{location}/videos')
     .mock(Response(
         status_code=200,
         content=_videos)))

    (respx_mock
     .get(f'{location}/videos')
     .mock(Response(
         status_code=200,
         content=_videos)))


    videos = (
        social.videos_block(
            {'id': 'mocked'}))

    video = (
        social.video_block('mocked'))


    sample_path = (
        f'{SAMPLES}/videos'
        '/dumped.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in videos],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in videos])

    assert expect == sample


    assert video == videos[0]



@mark.asyncio
async def test_YouTube_videos_async(
    social: YouTube,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _videos = read_text(
        f'{SAMPLES}/videos'
        '/source.json')

    location = (
        'https://www.googleapis.com'
        '/youtube/v3')


    (respx_mock
     .get(f'{location}/videos')
     .mock(Response(
         status_code=200,
         content=_videos)))

    (respx_mock
     .get(f'{location}/videos')
     .mock(Response(
         status_code=200,
         content=_videos)))


    videos = await (
        social.videos_async(
            {'id': 'mocked'}))

    video = await (
        social.video_async('mocked'))


    sample_path = (
        f'{SAMPLES}/videos'
        '/dumped.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in videos],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in videos])

    assert expect == sample


    assert video == videos[0]



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .helpers import EVENTS
from .helpers import IRCClientSocket
from .helpers import RVENTS
from .helpers import SVENTS
from .helpers import client_ircsock



__all__ = [
    'IRCClientSocket',
    'client_ircsock',
    'EVENTS',
    'RVENTS',
    'SVENTS']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from httpx import Response

from .models import RedditListing
from ..utils import HTTPClient
from ..utils.http import _HTTPAUTH
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import RedditParams



class Reddit:
    """
    Interact with the cloud service API with various methods.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'RedditParams'
    __client: HTTPClient

    __token: Optional[str]


    def __init__(
        self,
        params: 'RedditParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client

        self.__token = None


    @property
    def params(
        self,
    ) -> 'RedditParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    @property
    def token(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__token


    def request_token_block(
        self,
    ) -> str:
        """
        Establish new session obtaining token for authorization.

        :returns: Access token used with authenticated requests.
        """

        if self.__token is not None:
            return self.__token

        _params = self.__params
        username = _params.username
        password = _params.password
        client = _params.client
        secret = _params.secret

        request = self.request_block

        payload = {
            'grant_type': 'password',
            'username': username,
            'password': password}

        response = request(
            method='post',
            path='api/v1/access_token',
            data=payload,
            httpauth=(client, secret))

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)

        self.__token = (
            fetched['access_token'])

        return self.__token


    async def request_token_async(
        self,
    ) -> str:
        """
        Establish new session obtaining token for authorization.

        :returns: Access token used with authenticated requests.
        """

        if self.__token is not None:
            await asyncio.sleep(0)
            return self.__token

        _params = self.__params
        username = _params.username
        password = _params.password
        client = _params.client
        secret = _params.secret

        request = self.request_async

        payload = {
            'grant_type': 'password',
            'username': username,
            'password': password}

        response = await request(
            method='post',
            path='api/v1/access_token',
            data=payload,
            httpauth=(client, secret))

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)

        self.__token = (
            fetched['access_token'])

        return self.__token


    def request_block(
        self,
        method: Literal['get', 'post'],
        path: str,
        params: Optional[_PAYLOAD] = None,
        data: Optional[_PAYLOAD] = None,
        *,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param data: Optional dict payload included in request.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        server = 'www.reddit.com'
        client = self.__client

        request = client.request_block


        token = self.__token
        tokey = 'Authorization'

        _params = self.__params
        agent = _params.useragent

        headers = {
            'User-Agent': agent}

        if token is not None:
            headers[tokey] = token
            server = 'oauth.reddit.com'


        location = (
            f'https://{server}/{path}')

        return request(
            method=method,
            location=location,
            params=params,
            data=data,
            headers=headers,
            httpauth=httpauth)


    async def request_async(
        self,
        method: Literal['get', 'post'],
        path: str,
        params: Optional[_PAYLOAD] = None,
        data: Optional[_PAYLOAD] = None,
        *,
        httpauth: Optional[_HTTPAUTH] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param data: Optional dict payload included in request.
        :param httpauth: Optional information for authentication.
        :returns: Response from upstream request to the server.
        """

        server = 'www.reddit.com'
        client = self.__client

        request = client.request_async


        token = self.__token
        tokey = 'Authorization'

        _params = self.__params
        agent = _params.useragent

        headers = {
            'User-Agent': agent}

        if token is not None:
            headers[tokey] = token
            server = 'oauth.reddit.com'


        location = (
            f'https://{server}/{path}')

        return await request(
            method=method,
            location=location,
            params=params,
            data=data,
            headers=headers,
            httpauth=httpauth)


    def listing_block(
        self,
        unique: str,
    ) -> RedditListing:
        """
        Return the specific content within the social platform.

        :param unique: Unique identifier within social platform.
        :returns: Specific content within the social platform.
        """

        request = self.request_block

        if self.__token is None:
            self.request_token_block()


        def _request() -> Response:

            return request(
                method='get',
                path=f'comments/{unique}.json')


        response = _request()

        if response.status_code == 401:

            self.__token = None

            self.request_token_block()

            response = _request()

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, list)


        source = fetched[0]['data']
        children = source['children']

        return RedditListing(
            **children[0]['data'])


    async def listing_async(
        self,
        unique: str,
    ) -> RedditListing:
        """
        Return the specific content within the social platform.

        :param unique: Unique identifier within social platform.
        :returns: Specific content within the social platform.
        """

        request = self.request_async

        if self.__token is None:
            await self.request_token_async()


        async def _request() -> Response:

            return await request(
                method='get',
                path=f'comments/{unique}.json')


        response = await _request()

        if response.status_code == 401:

            self.__token = None

            await self.request_token_async()

            response = await _request()

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, list)


        source = fetched[0]['data']
        children = source['children']

        return RedditListing(
            **children[0]['data'])


    def latest(
        # NOCVR
        self,
        subred: str,
        params: Optional[_PAYLOAD] = None,
    ) -> list[RedditListing]:
        """
        Return the new items within the provided subreddit path.

        :param subred: Path to subreddit containing the content.
        :param params: Optional parameters included in request.
        :returns: New items within the provided subreddit path.
        """

        return self.latest_block(subred, params)


    def latest_block(
        self,
        subred: str,
        params: Optional[_PAYLOAD] = None,
    ) -> list[RedditListing]:
        """
        Return the new items within the provided subreddit path.

        :param subred: Path to subreddit containing the content.
        :param params: Optional parameters included in request.
        :returns: New items within the provided subreddit path.
        """

        params = dict(params or {})

        request = self.request_block

        if self.__token is None:
            self.request_token_block()


        def _request() -> Response:

            return request(
                method='get',
                path=f'r/{subred}/new.json')


        response = _request()

        if response.status_code == 401:

            self.__token = None

            self.request_token_block()

            response = _request()

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        source = fetched['data']
        children = source['children']

        return [
            RedditListing(**x['data'])
            for x in children]


    async def latest_async(
        self,
        subred: str,
        params: Optional[_PAYLOAD] = None,
    ) -> list[RedditListing]:
        """
        Return the new items within the provided subreddit path.

        :param subred: Path to subreddit containing the content.
        :param params: Optional parameters included in request.
        :returns: New items within the provided subreddit path.
        """

        params = dict(params or {})

        request = self.request_async

        if self.__token is None:
            await self.request_token_async()


        async def _request() -> Response:

            return await request(
                method='get',
                path=f'r/{subred}/new.json')


        response = await _request()

        if response.status_code == 401:

            self.__token = None

            await self.request_token_async()

            response = await _request()

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        source = fetched['data']
        children = source['children']

        return [
            RedditListing(**x['data'])
            for x in children]



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from typing import AsyncIterator
from typing import Iterator
from unittest.mock import AsyncMock
from unittest.mock import patch

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from httpx import AsyncByteStream
from httpx import Response
from httpx import SyncByteStream

from pytest import fixture
from pytest import mark

from respx import MockRouter

from ..http import HTTPClient



@fixture
def client() -> HTTPClient:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    return HTTPClient()



def test_HTTPClient(
    client: HTTPClient,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting with server.
    """


    attrs = lattrs(client)

    assert attrs == [
        '_HTTPClient__timeout',
        '_HTTPClient__headers',
        '_HTTPClient__verify',
        '_HTTPClient__capem',
        '_HTTPClient__httpauth',
        '_HTTPClient__retry',
        '_HTTPClient__backoff',
        '_HTTPClient__states',
        '_HTTPClient__client_block',
        '_HTTPClient__client_async']


    assert inrepr(
        'http.HTTPClient object',
        client)

    assert isinstance(
        hash(client), int)

    assert instr(
        'http.HTTPClient object',
        client)


    assert client.timeout == 30

    assert not client.headers

    assert client.verify

    assert not client.capem

    assert not client.httpauth

    assert client.retry == 3

    assert client.backoff == 3.0

    assert client.states == {429}

    assert client.client_block

    assert client.client_async



def test_HTTPClient_request_block(
    client: HTTPClient,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting with server.
    """

    patched = patch(
        'httpx.Client.request')

    request = client.request_block

    with patched as mocker:

        mocker.side_effect = [
            Response(429),
            Response(200)]

        response = request(
            'get', 'https://enasis.net')

        status = response.status_code

        assert status == 200

        assert mocker.call_count == 2



@mark.asyncio
async def test_HTTPClient_request_async(
    client: HTTPClient,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting with server.
    """

    patched = patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock)

    request = client.request_async

    with patched as mocker:

        mocker.side_effect = [
            Response(429),
            Response(200)]

        response = await request(
            'get', 'https://enasis.net')

        status = response.status_code

        assert status == 200

        assert mocker.call_count == 2



def test_HTTPClient_stream_block(
    client: HTTPClient,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting with server.
    :param respx_mock: Object for mocking request operation.
    """


    location = (
        'https://192.168.1.10'
        '/eventstream/clip/v2')


    class ByteStream(SyncByteStream):

        def __iter__(
            self,
        ) -> Iterator[bytes]:

            chunks = [
                b'stream content\n',
                b'stream content\n']

            yield from chunks


    streamer = ByteStream()

    (respx_mock
     .get(location)
     .mock(Response(
         status_code=200,
         stream=streamer)))


    request = client.stream_block

    stream = request(
        'get', location)


    chunks: list[str] = []

    for chunk in stream:
        chunks.append(chunk)

    assert chunks == [
        'stream content',
        'stream content']



@mark.asyncio
async def test_HTTPClient_stream_async(
    client: HTTPClient,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting with server.
    :param respx_mock: Object for mocking request operation.
    """


    location = (
        'https://192.168.1.10'
        '/eventstream/clip/v2')


    class ByteStream(AsyncByteStream):

        async def __aiter__(
            self,
        ) -> AsyncIterator[bytes]:

            chunks = [
                b'stream content\n',
                b'stream content\n']

            await asyncio.sleep(0)

            for chunk in chunks:

                yield chunk

                await asyncio.sleep(0)

            await asyncio.sleep(0)


    streamer = ByteStream()

    (respx_mock
     .get(location)
     .mock(Response(
         status_code=200,
         stream=streamer)))


    request = client.stream_async

    stream = request(
        'get', location)


    chunks: list[str] = []

    async for chunk in stream:
        chunks.append(chunk)

    assert chunks == [
        'stream content',
        'stream content']


    await asyncio.sleep(0)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from json import loads
from queue import Queue
from threading import Event
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Timer
from encommon.types import DictStrAny
from encommon.types import NCNone
from encommon.types import sort_dict

from httpx import Response

from websockets.exceptions import ConnectionClosedOK
from websockets.sync.client import ClientConnection
from websockets.sync.client import connect

from .models import ClientEvent
from ..utils import HTTPClient
from ..utils import dumlog
from ..utils.http import _METHODS
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import ClientParams



PING = {
    'action': 'ping',
    'seq': 1}



class Client:
    """
    Establish and maintain connection with the chat service.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'ClientParams'
    __logger: Callable[..., None]

    __client: HTTPClient
    __socket: Optional[ClientConnection]
    __conned: Event
    __exited: Event
    __mynick: Optional[tuple[str, str]]
    __lsnick: Optional[tuple[str, str]]

    __mqueue: Queue[ClientEvent]
    __cancel: Event


    def __init__(
        self,
        params: 'ClientParams',
        logger: Optional[Callable[..., None]] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params
        self.__logger = (
            logger or dumlog)

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client
        self.__socket = None
        self.__conned = Event()
        self.__exited = Event()
        self.__mynick = None
        self.__lsnick = None

        self.__mqueue = Queue(
            params.queue_size)

        self.__cancel = Event()


    @property
    def params(
        self,
    ) -> 'ClientParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def connected(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return (
            not self.__exited.is_set()
            and self.__conned.is_set())


    @property
    def nickname(
        self,
    ) -> Optional[tuple[str, str]]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mynick or self.__lsnick


    @property
    def mqueue(
        self,
    ) -> Queue[ClientEvent]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mqueue


    @property
    def canceled(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return (
            self.__cancel.is_set()
            or self.__exited.is_set())


    def operate(
        self,
    ) -> None:
        """
        Operate the client and populate queue with the messages.
        """

        logger = self.__logger

        try:

            logger(item='initial')

            self.__socket = None
            self.__conned.clear()
            self.__exited.clear()
            self.__mynick = None

            self.__cancel.clear()

            logger(item='operate')

            self.__operate()

        finally:

            self.__socket = None
            self.__conned.clear()
            self.__exited.clear()
            self.__mynick = None

            self.__cancel.clear()

            logger(item='finish')


    def __operate(
        self,
    ) -> None:
        """
        Operate the client and populate queue with the messages.
        """

        logger = self.__logger


        self.__connect()

        socket = self.__socket

        assert socket is not None


        timer = Timer(
            30, start='min')


        self.__identify()


        while not self.canceled:

            receive = (
                self.socket_recv())

            if receive is not None:
                self.__event(receive)

            if timer.pause():
                continue  # NOCVR

            logger(item='ping')

            self.socket_send(PING)


        logger(item='close')

        socket.close(1000)


        if self.__exited.is_set():
            raise ConnectionError


    def __event(
        self,
        event: DictStrAny,
    ) -> None:
        """
        Operate the client and populate queue with the messages.

        :param event: Raw event received from the network peer.
        """

        logger = self.__logger
        mqueue = self.__mqueue

        tneve = event.get('event')

        model = ClientEvent


        if tneve == 'hello':

            logger(item='helo')


        object = model(
            self, event)

        mqueue.put(object)


    def stop(
        self,
    ) -> None:
        """
        Gracefully close the connection with the server socket.
        """

        logger = self.__logger

        logger(item='stop')

        self.__cancel.set()


    def __connect(
        self,
    ) -> None:
        """
        Establish the connection with the upstream using socket.
        """

        logger = self.__logger

        _params = self.__params
        server = _params.server

        logger(item='connect')

        socket = connect(
            f'wss://{server}/'
            'api/v4/websocket')

        self.__socket = socket

        self.__conned.set()
        self.__exited.clear()


    def __identify(
        self,
    ) -> None:
        """
        Identify the client once the connection is established.
        """

        logger = self.__logger
        request = self.request

        _params = self.__params
        token = _params.token

        logger(item='identify')

        action = (
            'authentication'
            '_challenge')

        data = {'token': token}

        self.socket_send({
            'seq': 1,
            'action': action,
            'data': data})


        response = request(
            'get', 'users/me')

        (response
         .raise_for_status())


        fetch = response.json()

        assert isinstance(fetch, dict)

        logger(
            item='whome',
            json=dumps(fetch))

        self.__mynick = (
            (fetch['username']
             .lstrip('@')),
            fetch['id'])

        self.__lsnick = (
            (fetch['username']
             .lstrip('@')),
            fetch['id'])


    def socket_send(
        self,
        send: DictStrAny,
    ) -> None:
        """
        Transmit provided content through the socket connection.

        :param send: Content which will be sent through socket.
        """

        logger = self.__logger
        exited = self.__exited
        socket = self.__socket

        if socket is None:
            return NCNone

        transmit = dumps(send)

        logger(
            item='transmit',
            value=transmit)

        try:
            socket.send(transmit)

        except ConnectionClosedOK:
            exited.set()
            return None


    def socket_recv(  # noqa: CFQ004
        self,
    ) -> Optional[DictStrAny]:
        """
        Return the content received from the socket connection.

        :returns: Content received from the socket connection.
        """

        logger = self.__logger
        exited = self.__exited
        socket = self.__socket

        if socket is None:
            return NCNone


        try:

            recv = socket.recv(1)

            logger(
                item='receive',
                value=recv)

        except TimeoutError:
            return None

        except ConnectionClosedOK:
            exited.set()
            return None


        event = loads(recv)

        assert isinstance(event, dict)

        type = event.get('event')

        if type == 'discon':
            exited.set()

        return sort_dict(event)


    def request(
        self,
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})
        json = dict(json or {})

        logger = self.__logger
        client = self.__client

        _params = self.__params
        server = _params.server
        port = _params.port
        token = _params.token

        address = f'{server}:{port}'
        tokey = 'Authorization'
        content = 'application/json'

        headers = {
            tokey: f'Bearer {token}',
            'Content-Type': content}

        location = (
            f'https://{address}'
            f'/api/v4/{path}')

        request = client.request_block

        logger(
            item='request',
            method=method,
            path=path,
            params=params,
            json=(
                dumps(json)
                if len(json) >= 1
                else None))

        return request(
            method=method,
            location=location,
            params=params,
            headers=headers,
            json=json,
            timeout=timeout)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture

from respx import MockRouter

from . import SAMPLES
from ..params import RouterParams
from ..router import Router



@fixture
def router() -> Router:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = RouterParams(
        server='192.168.1.1',
        username='mocked',
        password='mocked')

    return Router(params)



def test_Router(
    router: Router,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param router: Class instance for connecting to service.
    """


    attrs = lattrs(router)

    assert attrs == [
        '_Router__params',
        '_Router__client']


    assert inrepr(
        'router.Router object',
        router)

    assert isinstance(
        hash(router), int)

    assert instr(
        'router.Router object',
        router)


    assert router.params

    assert router.client



def test_Router_request(
    router: Router,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param router: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    source = read_text(
        SAMPLES / 'source.json')

    source = read_sample(
        sample=source)


    (respx_mock
     .get(
         'https://192.168.1.1'
         '/proxy/network/api/s'
         '/default/rest/user')
     .mock(side_effect=[
         Response(401),
         Response(
             status_code=200,
             content=source)]))

    (respx_mock
     .post(
         'https://192.168.1.1'
         '/api/auth/login')
     .mock(Response(200)))


    request = router.reqroxy

    response = request(
        method='get',
        path='rest/user')

    response.raise_for_status()

    fetched = response.json()


    sample_path = (
        SAMPLES / 'dumped.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=fetched)

    expect = prep_sample(
        content=fetched)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .dummy import dumlog
from .http import HTTPClient



__all__ = [
    'HTTPClient',
    'dumlog']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .discord.test import DSCClientSocket
from .discord.test import client_dscsock
from .irc.test import IRCClientSocket
from .irc.test import client_ircsock
from .mattermost.test import MTMClientSocket
from .mattermost.test import client_mtmsock



__all__ = [
    'IRCClientSocket',
    'client_ircsock',
    'DSCClientSocket',
    'client_dscsock',
    'MTMClientSocket',
    'client_mtmsock']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field



class BridgeParams(BaseModel, extra='forbid'):
    """
    Process and validate the class configuration parameters.
    """

    server: Annotated[
        str,
        Field(...,
              description='Server address for connection',
              min_length=1)]

    timeout: Annotated[
        int,
        Field(30,
              description='Timeout connecting to server',
              ge=1, le=300)]

    token: Annotated[
        str,
        Field(...,
              description='Parameter for the integration',
              min_length=1)]

    ssl_verify: Annotated[
        bool,
        Field(True,
              description='Verify the ceritifcate valid')]

    ssl_capem: Annotated[
        Optional[str],
        Field(None,
              description='Verify the ceritifcate valid',
              min_length=1)]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Literal
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field



class ClientParams(BaseModel, extra='forbid'):
    """
    Process and validate the class configuration parameters.
    """

    server: Annotated[
        str,
        Field(...,
              description='Server address for connection',
              min_length=1)]

    port: Annotated[
        int,
        Field(6697,
              description='Server address for connection',
              ge=1, le=65535)]

    timeout: Annotated[
        int,
        Field(30,
              description='Timeout connecting to server',
              ge=1, le=300)]

    operate: Annotated[
        Literal['normal', 'service'],
        Field('normal',
              description='Method for server connection')]

    nickname: Annotated[
        str,
        Field('ircbot',
              description='Parameter for the integration',
              min_length=1)]

    username: Annotated[
        str,
        Field('ircbot',
              description='Parameter for the integration',
              min_length=1)]

    realname: Annotated[
        str,
        Field('Chatting Robie',
              description='Parameter for the integration',
              min_length=1)]

    password: Annotated[
        Optional[str],
        Field(None,
              description='Parameter for the integration',
              min_length=1)]

    servername: Annotated[
        str,
        Field('services.invalid',
              description='Parameter for the integration',
              min_length=1)]

    serverid: Annotated[
        str,
        Field('42X',
              description='Unique identifier for services',
              min_length=1)]

    ssl_enable: Annotated[
        bool,
        Field(True,
              description='Enable connection encryption')]

    ssl_verify: Annotated[
        bool,
        Field(True,
              description='Verify the ceritifcate valid')]

    queue_size: Annotated[
        int,
        Field(10000,
              description='Maximum size for queued events',
              ge=1000, le=1000000)]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture

from respx import MockRouter

from . import SAMPLES
from ..bridge import Bridge
from ..params import BridgeParams



@fixture
def bridge() -> Bridge:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = BridgeParams(
        server='192.168.1.11',
        appid=69,
        token='mocked')

    return Bridge(params)



def test_Bridge(
    bridge: Bridge,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    """


    attrs = lattrs(bridge)

    assert attrs == [
        '_Bridge__params',
        '_Bridge__client']


    assert inrepr(
        'bridge.Bridge object',
        bridge)

    assert isinstance(
        hash(bridge), int)

    assert instr(
        'bridge.Bridge object',
        bridge)


    assert bridge.params

    assert bridge.client



def test_Bridge_request(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    source = read_text(
        SAMPLES / 'source.json')

    source = read_sample(
        sample=source)


    (respx_mock
     .get(
         'https://192.168.1.11'
         '/apps/api/69'
         '/devices/all')
     .mock(Response(
         status_code=200,
         content=source)))


    request = bridge.request

    response = request(
        method='get',
        path='devices/all')

    response.raise_for_status()

    fetched = response.json()


    sample_path = (
        SAMPLES / 'dumped.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=fetched)

    expect = prep_sample(
        content=fetched)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import IGNORECASE
from re import escape as re_escape
from re import search as re_search
from typing import Annotated
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import BaseModel
from encommon.types import DictStrAny
from encommon.types import NCNone

from pydantic import Field

if TYPE_CHECKING:
    from .client import Client



KINDS = Literal[
    'event',
    'chanmsg',
    'privmsg']

MESSAGE = [
    'chanmsg',
    'privmsg']



class ClientEvent(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream server.
    """

    type: Annotated[
        Optional[str],
        Field(None,
              description='Type of event that occurred',
              min_length=1)]

    opcode: Annotated[
        int,
        Field(...,
              description='Type of operation performed',
              ge=0)]

    data: Annotated[
        Optional[DictStrAny],
        Field(None,
              description='Payload with the event data',
              min_length=1)]

    seqno: Annotated[
        Optional[int],
        Field(None,
              description='Event number within squence',
              ge=0)]

    original: Annotated[
        DictStrAny,
        Field(...,
              description='Original received from server',
              min_length=1)]

    kind: Annotated[
        KINDS,
        Field('event',
              description='Dynamic field parsed from event')]

    isme: Annotated[
        bool,
        Field(False,
              description='Indicates message is from client')]

    hasme: Annotated[
        bool,
        Field(False,
              description='Indicates message mentions client')]

    whome: Annotated[
        Optional[tuple[str, str]],
        Field(None,
              description='Current nickname of when received',
              min_length=1)]

    author: Annotated[
        Optional[tuple[str, str]],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    recipient: Annotated[
        Optional[tuple[Optional[str], str]],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    message: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]


    def __init__(
        self,
        /,
        client: 'Client',
        event: DictStrAny,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        data: DictStrAny = {
            'original': event}


        type = event.get('t')
        opcode = event.get('op')
        _data = event.get('d')
        seqno = event.get('s')


        if type is not None:
            data['type'] = type

        if opcode is not None:
            data['opcode'] = opcode

        if (isinstance(_data, dict)
                and len(_data) >= 1):
            data['data'] = _data

        if seqno is not None:
            data['seqno'] = seqno


        super().__init__(**data)

        self.__set_kind()
        self.__set_author()
        self.__set_recipient()
        self.__set_message()
        self.__set_isme(client)
        self.__set_hasme(client)
        self.__set_whome(client)


    def __set_kind(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        type = self.type
        data = self.data

        kind: KINDS = 'event'


        message = [
            'MESSAGE_CREATE',
            'MESSAGE_UPDATE']

        if type in message:

            assert data is not None

            guild = (
                data.get('guild_id'))

            kind = (
                'chanmsg'
                if guild is not None
                else 'privmsg')


        self.kind = kind


    def __set_author(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        dscuser = (
            data.get('author'))

        if dscuser is None:
            return NCNone

        unique = dscuser['id']
        name = dscuser['username']

        self.author = (
            name, unique)


    def __set_recipient(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        guild = (
            data.get('guild_id'))

        channel = (
            data.get('channel_id'))

        if channel is None:
            return NCNone

        self.recipient = (
            guild, channel)


    def __set_message(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        content = (
            data.get('content'))

        self.message = content


    def __set_isme(
        self,
        client: 'Client',
    ) -> None:
        """
        Update the value for the attribute from class instance.

        :param client: Class instance for connecting to service.
        """

        mynick = client.nickname
        author = self.author

        isme: bool = False

        if mynick and author:

            mine = mynick[1]
            them = author[1]

            isme = mine == them

        self.isme = isme


    def __set_hasme(
        self,
        client: 'Client',
    ) -> None:
        """
        Update the value for the attribute from class instance.

        :param client: Class instance for connecting to service.
        """

        mynick = client.nickname
        message = self.message

        if mynick is None:
            return None

        if message is None:
            return None

        needle = (
            re_escape(mynick[0]),
            re_escape(mynick[1]))

        pattern = (
            rf'\b({needle[0]})'
            rf'|({needle[1]})\b')

        mentions = [
            x['id'] for x in
            (self.original
             .get('mentions', {})
             .values())]

        self.hasme = any([

            mynick[1] in mentions,

            re_search(
                pattern,
                message,
                IGNORECASE)])


    def __set_whome(
        self,
        client: 'Client',
    ) -> None:
        """
        Update the value for the attribute from class instance.

        :param client: Class instance for connecting to service.
        """

        self.whome = client.nickname
